# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines

import abc
import datetime
import enum
import json
import math
import types

import ipaddress

try:
    from collections.abc import MutableSequence  # only works on python 3.3+
except ImportError:  # pragma: no cover
    from collections import MutableSequence  # pylint: disable=deprecated-class

from collections import OrderedDict

import attr
import urllib3

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.grade import Gradeable
from cryptodatahub.common.types import CryptoDataEnumCodedBase, CryptoDataParamsBase

from cryptoparser.common.parse import (
    ComposerBinary,
    ComposerText,
    ParsableBase,
    ParsableBaseNoABC,
    ParserBinary,
    ParserText,
)
from cryptoparser.common.exception import NotEnoughData, TooMuchData, InvalidType
from cryptoparser.common.utils import bytes_to_hex_string


def _default(
        self,  # pylint: disable=unused-argument
        obj
):
    result = Serializable._json_traverse(obj, Serializable._json_result)  # pylint: disable=protected-access
    return result


_default.default = json.JSONEncoder().default
json.JSONEncoder.default = _default


class SerializableTextEncoder():
    def __call__(self, obj, level):
        if isinstance(obj, str):
            string_result = obj
        else:
            string_result = str(obj)

        return False, string_result


class Serializable():  # pylint: disable=too-few-public-methods
    _MARKDOWN_RESULT_STRING_CLASSES = (
        ipaddress.IPv4Network,
        ipaddress.IPv6Network,
        urllib3.util.url.Url,
    )
    post_text_encoder = SerializableTextEncoder()

    @staticmethod
    def _filter_out_non_human_friendly(obj, dict_value, human_friendly_only):
        if not attr.has(type(obj)) or not human_friendly_only:
            return dict_value

        fields_dict = attr.fields_dict(type(obj))
        dict_value = OrderedDict([
            (name, value)
            for name, value in dict_value.items()
            if name not in fields_dict or fields_dict[name].metadata.get('human_friendly', True)
        ])

        return dict_value

    @staticmethod
    def _get_ordered_dict(dict_value, human_friendly_only=False):
        if attr.has(type(dict_value)):
            obj = dict_value
            dict_value = OrderedDict([
                (name, getattr(dict_value, name))
                for name, field in attr.fields_dict(type(dict_value)).items()
                if not name.startswith('_')
            ])
            dict_value = Serializable._filter_out_non_human_friendly(obj, dict_value, human_friendly_only)
            keys = dict_value.keys()
        elif isinstance(dict_value, OrderedDict):
            keys = dict_value.keys()
        elif isinstance(dict_value, dict):
            if all(isinstance(key, enum.Enum) for key in dict_value.keys()):
                keys = sorted(dict_value.keys(), key=lambda key: key.name)
            else:
                keys = sorted(dict_value.keys())
        elif hasattr(dict_value, '__dict__'):
            dict_value = dict_value.__dict__
            keys = sorted(filter(lambda key: not key.startswith('_'), dict_value.keys()))

        result = OrderedDict([
            (key, dict_value[key])
            for key in keys
        ])

        return result

    @staticmethod
    def _json_result(obj):
        if isinstance(obj, enum.Enum):
            if isinstance(obj.value, CryptoDataParamsBase):
                result = obj.name
            else:
                result = {obj.name: obj.value}
        elif isinstance(obj, (str, int, float, bool, )) or obj is None:
            result = obj
        elif isinstance(obj, (bytes, bytearray)):
            result = bytes_to_hex_string(obj, separator=':', lowercase=False)
        else:
            result = str(obj)

        return result

    @staticmethod
    def _json_traverse(obj, result_func):
        if isinstance(obj, enum.Enum):
            result = result_func(obj)
        elif hasattr(obj, '_asdict'):
            result = Serializable._json_traverse(obj._asdict(), result_func)
        elif isinstance(obj, dict) or attr.has(type(obj)):
            result = OrderedDict([
                (
                    key.name if isinstance(key, enum.Enum) else Serializable._json_result(key),
                    Serializable._json_traverse(value, result_func)
                )
                for key, value in Serializable._get_ordered_dict(obj).items()
            ])
        elif hasattr(obj, '__dict__'):
            result = Serializable._json_traverse(obj.__dict__, result_func)
        elif isinstance(obj, (list, tuple, frozenset, set)):
            result = [Serializable._json_traverse(item, result_func) for item in obj]
        else:
            result = result_func(obj)

        return result

    @staticmethod
    def _markdown_indent_from_level(level):
        return 4 * level * ' '

    @classmethod
    def _markdown_human_readable_names(cls, obj, dict_value):
        name_dict = {}
        fields_dict = attr.fields_dict(type(obj)) if attr.has(type(obj)) else {}
        for name in dict_value:
            if isinstance(name, str):
                if name in fields_dict and 'human_readable_name' in fields_dict[name].metadata:
                    human_readable_name = fields_dict[name].metadata['human_readable_name']
                else:
                    human_readable_name = ' '.join(name.split('_')).title()
            else:
                post_text_encoder = cls.post_text_encoder
                cls.post_text_encoder = SerializableTextEncoder()
                _, human_readable_name = cls._markdown_result(name)
                cls.post_text_encoder = post_text_encoder

            name_dict[name] = human_readable_name

        return name_dict

    @classmethod
    def _markdown_result_complex(cls, obj, level=0):
        indent = Serializable._markdown_indent_from_level(level)

        if hasattr(obj, '_asdict'):
            dict_value = obj._asdict()
            if not isinstance(dict_value, dict):
                return False, dict_value

            dict_value = Serializable._filter_out_non_human_friendly(obj, dict_value, human_friendly_only=True)
        else:
            dict_value = Serializable._get_ordered_dict(obj, human_friendly_only=True)

        result = ''
        name_dict = cls._markdown_human_readable_names(obj, dict_value)
        for name, value in dict_value.items():
            result += f'{indent}* {name_dict[name]}'
            multiline, markdnow_result = cls._markdown_result(value, level + 1)
            if multiline:
                result += f':\n{markdnow_result}'
            else:
                result += f': {markdnow_result}\n'

        if not result:
            return False, '-'

        return True, result

    @classmethod
    def _markdown_result_list(cls, obj, level=0):
        if not obj:
            return False, '-'

        indent = Serializable._markdown_indent_from_level(level)

        result = ''
        for index, item in enumerate(obj):
            multiline, markdnow_result = cls._markdown_result(item, level + 1)
            separator = '\n' if multiline else ' '
            newline = '' if multiline else '\n'
            result += f'{indent}{index + 1}.{separator}{markdnow_result}{newline}'

        return True, result

    @staticmethod
    def _markdown_is_directly_printable(obj):
        return not isinstance(obj, enum.Enum) and isinstance(obj, (str, int, float, ))

    @classmethod
    def _markdown_result(cls, obj, level=0):  # pylint: disable=too-many-branches,too-many-return-statements
        if obj is None:
            result = cls.post_text_encoder('n/a', level)
        elif isinstance(obj, bool):
            result = cls.post_text_encoder('yes' if obj else 'no', level)
        elif Serializable._markdown_is_directly_printable(obj):
            result = cls.post_text_encoder(obj, level)
        elif isinstance(obj, Gradeable):
            result = cls.post_text_encoder(obj, level)
        elif isinstance(obj, Serializable):
            result = obj._as_markdown(level)  # pylint: disable=protected-access
        elif isinstance(obj, enum.Enum):
            if isinstance(obj.value, Serializable):
                return obj.value._as_markdown(level)  # pylint: disable=protected-access
            if isinstance(obj.value, CryptoDataParamsBase):
                return cls.post_text_encoder(obj.value, level)

            return cls.post_text_encoder(obj.name, level)
        elif isinstance(obj, cls._MARKDOWN_RESULT_STRING_CLASSES):
            return False, str(obj)
        elif isinstance(obj, datetime.timedelta):
            return False, str(obj.seconds)
        elif isinstance(obj, CryptoDataParamsBase) and hasattr(obj, '__str__'):
            return False, str(obj)
        elif attr.has(type(obj)):
            result = cls._markdown_result_complex(obj, level)
        elif hasattr(obj, '_asdict'):
            result = cls._markdown_result(obj._asdict(), level)
        elif hasattr(obj, '__dict__') or isinstance(obj, dict):
            result = cls._markdown_result_complex(obj, level)
        elif isinstance(obj, (list, tuple, frozenset, set, ArrayBase)):
            result = cls._markdown_result_list(obj, level)
        elif isinstance(obj, (bytes, bytearray)):
            result = cls.post_text_encoder(bytes_to_hex_string(obj, separator=':', lowercase=False), level)
        else:
            result = cls.post_text_encoder(obj, level)

        return result

    def _asdict(self):
        return Serializable._get_ordered_dict(self)

    def as_json(self):
        return json.dumps(self)

    def _as_markdown(self, level):
        return self._markdown_result_complex(self, level)

    def as_markdown(self):
        _, result = self._as_markdown(0)
        return result


@attr.s
class VariantParsableBase(ParsableBase):
    variant = attr.ib()

    _REGISTERED_VARIANTS = OrderedDict()

    @classmethod
    @abc.abstractmethod
    def _get_variants(cls):
        raise NotImplementedError()

    @variant.validator
    def _validator_variant(self, _, value):
        for variant_type in self._get_variant_types():
            if issubclass(variant_type, NByteEnumParsable):
                variant_type = variant_type.get_enum_class()

            if isinstance(value, variant_type):
                break
        else:
            raise InvalidValue(value, VariantParsable)

    @classmethod
    def _get_variant_types(cls):
        variant_types = []

        for variant_type_list in list(cls._get_variants().values()) + list(cls._get_registered_variants().values()):
            variant_types.extend(variant_type_list)

        return variant_types

    @classmethod
    def _get_registered_variants(cls):
        if cls not in cls._REGISTERED_VARIANTS:
            cls._REGISTERED_VARIANTS[cls] = OrderedDict()

        return cls._REGISTERED_VARIANTS[cls]

    @classmethod
    def register_variant_parser(cls, variant_tag, parsable_class):
        registered_variants = cls._get_registered_variants()
        if variant_tag not in registered_variants:
            registered_variants[variant_tag] = []

        registered_variants[variant_tag].append(parsable_class)

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    def compose(self):
        return self.variant.compose()


class VariantParsable(VariantParsableBase):
    @classmethod
    @abc.abstractmethod
    def _get_variants(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        for variant_parser in cls._get_variant_types():
            try:
                parsed_object, parsed_length = variant_parser.parse_immutable(parsable)
                return parsed_object, parsed_length
            except InvalidType:
                pass

        raise InvalidValue(parsable, cls)


class VariantParsableExact(VariantParsableBase):
    @classmethod
    @abc.abstractmethod
    def _get_variants(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        for variant_parser in cls._get_variant_types():
            try:
                parsed_object = variant_parser.parse_exact_size(parsable)
                return parsed_object, len(parsable)
            except (InvalidType, InvalidValue, TooMuchData):
                pass

        raise InvalidValue(parsable, cls)


@attr.s
class VectorParamBase():  # pylint: disable=too-few-public-methods
    min_byte_num = attr.ib(validator=attr.validators.instance_of(int))
    max_byte_num = attr.ib(validator=attr.validators.instance_of(int))
    item_num_size = attr.ib(init=False, validator=attr.validators.instance_of(int))

    def __attrs_post_init__(self):
        self.item_num_size = int(math.log(self.max_byte_num, 2) / 8) + 1

        attr.validate(self)

    @abc.abstractmethod
    def get_item_size(self, item):
        raise NotImplementedError()


@attr.s
class VectorParamNumeric(VectorParamBase):  # pylint: disable=too-few-public-methods
    item_size = attr.ib(validator=attr.validators.instance_of(int))
    numeric_class = attr.ib(default=int, validator=attr.validators.instance_of(type))

    def get_item_size(self, item):
        return self.item_size


@attr.s(init=False)
class OpaqueParam(VectorParamNumeric):  # pylint: disable=too-few-public-methods
    def __init__(self, min_byte_num, max_byte_num):
        super().__init__(min_byte_num, max_byte_num, 1)

    def get_item_size(self, item):
        return 1


@attr.s
class VectorParamString(VectorParamBase):  # pylint: disable=too-few-public-methods
    separator = attr.ib(validator=attr.validators.instance_of(str), default=',')
    encoding = attr.ib(validator=attr.validators.instance_of(str), default='ascii')
    item_class = attr.ib(validator=attr.validators.instance_of((type, types.FunctionType)), default=str)
    fallback_class = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of((type, types.FunctionType)))
    )

    def get_item_size(self, item):
        if isinstance(item, (ParsableBase, StringEnumParsable)):
            return len(item.compose())
        if isinstance(item, CryptoDataEnumCodedBase):
            return item.value.get_code_size()
        if isinstance(item, str):
            return len(item)

        raise NotImplementedError(type(item))


@attr.s
class VectorParamParsable(VectorParamBase):  # pylint: disable=too-few-public-methods
    item_class = attr.ib(validator=attr.validators.instance_of((type, types.FunctionType)))
    fallback_class = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of((type, types.FunctionType)))
    )

    def get_item_size(self, item):
        return len(item.compose())


@attr.s
class VectorParamEnumCodeNumeric(VectorParamBase):  # pylint: disable=too-few-public-methods
    item_class = attr.ib(validator=attr.validators.instance_of((type, types.FunctionType)))
    fallback_class = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of((type, types.FunctionType)))
    )

    def get_item_size(self, item):
        return self.fallback_class.get_byte_num()


@attr.s
class VectorParamEnumCodeString(VectorParamBase):  # pylint: disable=too-few-public-methods
    item_class = attr.ib(validator=attr.validators.instance_of((type, types.FunctionType)))
    fallback_class = attr.ib(init=False, default=None)

    def get_item_size(self, item):
        return len(item.value.code)


@attr.s
class ArrayBase(ParsableBase, MutableSequence, Serializable):
    _items = attr.ib()
    _items_size = attr.ib(init=False, default=0)
    param = attr.ib(init=False, default=None)

    def __attrs_post_init__(self):
        items = self._items

        self.param = self.get_param()
        self._items = []

        for item in items:
            self._items.append(item)
            self._items_size += self.param.get_item_size(item)

        self._update_items_size(del_item=None, insert_item=None)

        attr.validate(self)

    def _update_items_size(self, del_item=None, insert_item=None):
        size_diff = 0

        if del_item is not None:
            size_diff -= self.param.get_item_size(del_item)
        if insert_item is not None:
            size_diff += self.param.get_item_size(insert_item)

        if self._items_size + size_diff < self.param.min_byte_num:
            raise NotEnoughData(self.param.min_byte_num)
        if self._items_size + size_diff > self.param.max_byte_num:
            raise TooMuchData(self.param.max_byte_num)

        self._items_size += size_diff

    @classmethod
    @abc.abstractmethod
    def get_param(cls):
        raise NotImplementedError()

    def __len__(self):
        return len(self._items)

    def __getitem__(self, index):
        return self._items[index]

    def __delitem__(self, index):
        self._update_items_size(del_item=self._items[index])

        del self._items[index]

    def __setitem__(self, index, value):
        self._update_items_size(del_item=self._items[index], insert_item=value)
        self._items[index] = value

    def __str__(self):
        return str(self._items)

    def insert(self, index, value):
        self._update_items_size(insert_item=value)

        self._items.insert(index, value)

    def append(self, value):
        self.insert(len(self._items), value)

    def _asdict(self):
        return self._items

    def _as_markdown(self, level):
        return self._markdown_result(self._asdict(), level)


class Vector(ArrayBase):
    @classmethod
    @abc.abstractmethod
    def get_param(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        vector_param = cls.get_param()
        parser = ParserBinary(parsable)

        parser.parse_numeric('item_byte_num', vector_param.item_num_size)

        item_byte_num = parser['item_byte_num']
        item_num = int(item_byte_num / vector_param.item_size)
        parser.parse_numeric_array('items', item_num, vector_param.item_size, vector_param.numeric_class)

        return cls(parser['items']), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(len(self._items) * self.param.item_size, self.param.item_num_size)
        composer.compose_numeric_array(self._items, self.param.item_size)

        return composer.composed_bytes


class VectorString(ArrayBase):
    @classmethod
    @abc.abstractmethod
    def get_param(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        vector_param = cls.get_param()

        header_parser = ParserBinary(parsable[:vector_param.item_num_size])

        header_parser.parse_numeric('item_byte_num', vector_param.item_num_size)

        if header_parser['item_byte_num'] == 0:
            return cls([]), header_parser.parsed_length

        body_parser = ParserText(
            parsable[vector_param.item_num_size:header_parser['item_byte_num'] + vector_param.item_num_size]
        )

        body_parser.parse_string_array(
            'items', vector_param.separator, vector_param.item_class, vector_param.fallback_class,
        )

        return cls(body_parser['items']), header_parser.parsed_length + body_parser.parsed_length

    def compose(self):
        vector_param = self.get_param()

        body_composer = ComposerText(vector_param.encoding)
        body_composer.compose_parsable_array(self._items, vector_param.separator, vector_param.fallback_class)

        header_composer = ComposerBinary()
        header_composer.compose_numeric(body_composer.composed_length, self.param.item_num_size)

        return header_composer.composed + body_composer.composed


class VectorParsable(ArrayBase):
    @classmethod
    @abc.abstractmethod
    def get_param(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        vector_param = cls.get_param()

        parser = ParserBinary(parsable)

        parser.parse_numeric('item_byte_num', vector_param.item_num_size)
        parser.parse_parsable_array(
            'items',
            items_size=parser['item_byte_num'],
            item_class=vector_param.item_class,
            fallback_class=vector_param.fallback_class
        )

        return cls(parser['items']), parser.parsed_length

    def compose(self):
        body_composer = ComposerBinary()
        body_composer.compose_parsable_array(self._items)

        header_composer = ComposerBinary()
        header_composer.compose_numeric(body_composer.composed_length, self.param.item_num_size)

        return header_composer.composed_bytes + body_composer.composed_bytes


class VectorEnumCodeNumeric(VectorParsable):
    @classmethod
    @abc.abstractmethod
    def get_param(cls):
        raise NotImplementedError()

    def compose(self):
        body_composer = ComposerBinary()

        for item in self:
            if isinstance(self.param.fallback_class, type) and isinstance(item, self.param.fallback_class):
                body_composer.compose_parsable(item)
            else:
                body_composer.compose_numeric_enum_coded(item)

        header_composer = ComposerBinary()
        header_composer.compose_numeric(body_composer.composed_length, self.param.item_num_size)

        return header_composer.composed_bytes + body_composer.composed_bytes


class VectorEnumCodeString(VectorParsable):
    @classmethod
    @abc.abstractmethod
    def get_param(cls):
        raise NotImplementedError()

    def compose(self):
        body_composer = ComposerBinary()

        item_size = self.get_param().item_class.get_param().item_num_size
        for item in self:
            body_composer.compose_string_enum_coded(item, item_size)

        header_composer = ComposerBinary()
        header_composer.compose_numeric(body_composer.composed_length, self.param.item_num_size)

        return header_composer.composed_bytes + body_composer.composed_bytes


class VectorParsableDerived(ArrayBase):
    @classmethod
    @abc.abstractmethod
    def get_param(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        vector_param = cls.get_param()

        parser = ParserBinary(parsable)

        parser.parse_numeric('item_byte_num', vector_param.item_num_size)
        parser.parse_parsable_derived_array(
            'items',
            items_size=parser['item_byte_num'],
            item_base_class=vector_param.item_class,
            fallback_class=vector_param.fallback_class
        )

        return cls(parser['items']), parser.parsed_length

    def compose(self):
        body_composer = ComposerBinary()
        body_composer.compose_parsable_array(self._items)

        header_composer = ComposerBinary()
        header_composer.compose_numeric(len(body_composer.composed_bytes), self.param.item_num_size)

        return header_composer.composed_bytes + body_composer.composed_bytes


class Opaque(ArrayBase):
    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('item_byte_num', cls.get_param().item_num_size)
        parser.parse_raw('items', parser['item_byte_num'])

        items = parser['items']
        return cls([ord(items[i:i + 1]) for i in range(len(items))]), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(len(self._items), self.get_param().item_num_size)
        composer.compose_numeric_array(self._items, 1)

        return composer.composed_bytes

    @classmethod
    @abc.abstractmethod
    def get_param(cls):
        raise NotImplementedError()


class NByteEnumParsable(ParsableBase):
    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('code', cls.get_byte_num())

        for enum_item in list(cls.get_enum_class()):
            if enum_item.value.code == parser['code']:
                return enum_item, cls.get_byte_num()

        raise InvalidValue(parser['code'], cls, 'code')

    @classmethod
    @abc.abstractmethod
    def get_byte_num(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_enum_class(cls):
        raise NotImplementedError()


class OneByteEnumParsable(NByteEnumParsable):
    @classmethod
    def get_byte_num(cls):
        return 1

    @classmethod
    @abc.abstractmethod
    def get_enum_class(cls):
        raise NotImplementedError()


class TwoByteEnumParsable(NByteEnumParsable):
    @classmethod
    def get_byte_num(cls):
        return 2

    @classmethod
    @abc.abstractmethod
    def get_enum_class(cls):
        raise NotImplementedError()


class ThreeByteEnumParsable(NByteEnumParsable):
    @classmethod
    def get_byte_num(cls):
        return 3

    @classmethod
    @abc.abstractmethod
    def get_enum_class(cls):
        raise NotImplementedError()


class FourByteEnumParsable(NByteEnumParsable):
    @classmethod
    def get_byte_num(cls):
        return 4

    @classmethod
    @abc.abstractmethod
    def get_enum_class(cls):
        raise NotImplementedError()


class NByteEnumComposer(enum.Enum):
    def __repr__(self):
        return self.__class__.__name__ + '.' + self.name

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(
            self.value.code,  # pylint: disable=no-member
            self.get_byte_num()
        )

        return composer.composed

    @classmethod
    @abc.abstractmethod
    def get_byte_num(cls):
        raise NotImplementedError()


class OneByteEnumComposer(NByteEnumComposer):
    @classmethod
    def get_byte_num(cls):
        return 1


class TwoByteEnumComposer(NByteEnumComposer):
    @classmethod
    def get_byte_num(cls):
        return 2


class ThreeByteEnumComposer(NByteEnumComposer):
    @classmethod
    def get_byte_num(cls):
        return 3


class FourByteEnumComposer(NByteEnumComposer):
    @classmethod
    def get_byte_num(cls):
        return 4


class StringEnumParsableBase(ParsableBaseNoABC):
    @classmethod
    @abc.abstractmethod
    def _code_eq(cls, item_code, parsed_code):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        enum_items = [
            enum_item
            for enum_item in cls  # pylint: disable=not-an-iterable
            if len(enum_item.value.code) <= len(parsable)
        ]
        enum_items.sort(key=lambda color: len(color.value.code), reverse=True)

        try:
            code = bytes(parsable).decode('ascii')
        except UnicodeDecodeError as e:
            raise InvalidValue(parsable, cls) from e

        for enum_item in enum_items:
            if cls._code_eq(enum_item.value.code, code[:len(enum_item.value.code)]):
                return enum_item, len(enum_item.value.code)

        raise InvalidValue(parsable, cls, 'code')

    def compose(self):
        return self._asdict().encode('ascii')

    def _asdict(self):
        return getattr(self, 'value').code


class StringEnumParsable(StringEnumParsableBase):
    @classmethod
    def _code_eq(cls, item_code, parsed_code):
        return item_code == parsed_code


class StringEnumCaseInsensitiveParsable(StringEnumParsableBase):
    @classmethod
    def _code_eq(cls, item_code, parsed_code):
        return item_code.lower() == parsed_code.lower()


class ProtocolVersionBase(Serializable, ParsableBase, metaclass=abc.ABCMeta):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def identifier(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def __str__(self):
        raise NotImplementedError()

    def _asdict(self):
        return self.identifier

    def _as_markdown(self, level):
        return self._markdown_result(str(self), level)


@attr.s
class ProtocolVersionMajorMinorBase(ProtocolVersionBase):
    _SIZE = 2

    major = attr.ib()
    minor = attr.ib()

    @classmethod
    def _parse_version_numbers(cls, parsable):
        if len(parsable) < cls._SIZE:
            raise NotEnoughData(bytes_needed=cls._SIZE - len(parsable))

        parser = ParserBinary(parsable)

        parser.parse_numeric('major', 1)
        parser.parse_numeric('minor', 1)

        return parser

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_version_numbers(parsable)

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.major, 1)
        composer.compose_numeric(self.minor, 1)

        return composer.composed_bytes

    @property
    def identifier(self):
        return f'{self.major}_{self.minor}'

    def __str__(self):
        return f'{self.major}.{self.minor}'


@attr.s
class ListParamParsable():  # pylint: disable=too-few-public-methods
    item_class = attr.ib(validator=attr.validators.instance_of(type))
    fallback_class = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(type)))
    separator_class = attr.ib(attr.validators.instance_of(ParsableBase))
    min_byte_num = attr.ib(init=False, default=0)
    max_byte_num = attr.ib(init=False, default=2 ** 16)
    item_num_size = attr.ib(init=False, default=0)

    def get_item_size(self, item):  # pylint: disable=no-self-use
        return len(item.compose())


class ListParsable(ArrayBase):
    @classmethod
    @abc.abstractmethod
    def get_param(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        list_param = cls.get_param()

        parser = ParserBinary(parsable)

        parser.parse_parsable_list(
            'items',
            item_class=list_param.item_class,
            fallback_class=list_param.fallback_class,
            separator_class=list_param.separator_class
        )

        return cls(parser['items']), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        separator = bytearray(self.get_param().separator_class().compose())
        composer.compose_parsable_array(self._items, separator)
        composer.compose_raw(separator)
        if self._items:
            composer.compose_raw(separator)

        return composer.composed_bytes


class OpaqueEnumParsable(Vector):
    @classmethod
    def _parse(cls, parsable):
        opaque, parsed_length = super(OpaqueEnumParsable, cls)._parse(parsable)
        code = b''.join([bytes((opaque_item,)) for opaque_item in opaque]).decode(cls.get_encoding())

        try:
            parsed_object = next(iter([
                enum_item
                for enum_item in cls.get_enum_class()
                if enum_item.value.code == code
            ]))
        except StopIteration as e:
            raise InvalidValue(code, cls) from e

        return parsed_object, parsed_length

    @classmethod
    @abc.abstractmethod
    def get_enum_class(cls):
        raise NotImplementedError()

    @classmethod
    def get_encoding(cls):
        return 'utf-8'


class OpaqueEnumComposer(enum.Enum):
    def __repr__(self):
        return self.__class__.__name__ + '.' + self.name

    def compose(self):
        composer = ComposerBinary()
        value = self.value.code.encode(self.get_encoding())  # pylint: disable=no-member

        composer.compose_bytes(value, 1)

        return composer.composed_bytes

    @classmethod
    def get_encoding(cls):
        return 'utf-8'


@attr.s
class NumericRangeParsableBase(ParsableBase, Serializable):
    value = attr.ib(validator=attr.validators.instance_of(int))

    @value.validator
    def _validator_variant(self, _, value):
        if value < self._get_value_min():
            raise InvalidValue(value, type(self))

        if value > self._get_value_max():
            raise InvalidValue(value, type(self))

    @classmethod
    @abc.abstractmethod
    def _get_value_min(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_value_max(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_value_length(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('value', cls._get_value_length())

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.value, self._get_value_length())

        return composer.composed

    def __str__(self):
        return str(self.value)

    def _as_markdown(self, level):
        return self._markdown_result(str(self), level)



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import NotEnoughData
from cryptoparser.tls.postgresql import SslRequest, Sync


class TestSslRequest(unittest.TestCase):
    def setUp(self):
        self.ssl_request = SslRequest()
        self.ssl_request_bytes = b'\x00\x00\x00\x08\x04\xd2\x16\x2f'

    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            SslRequest.parse_exact_size(b'')
        self.assertEqual(context_manager.exception.bytes_needed, SslRequest.MESSAGE_SIZE)

        with self.assertRaises(InvalidValue) as context_manager:
            SslRequest.parse_exact_size(b'\x00\x00\x00\x04\x01\x02\x03\x04')
        self.assertEqual(context_manager.exception.value, 4)

        with self.assertRaises(InvalidValue) as context_manager:
            SslRequest.parse_exact_size(b'\x00\x00\x00\x08\x01\x02\x03\x04')
        self.assertEqual(context_manager.exception.value, 0x01020304)

    def test_parse(self):
        SslRequest.parse_exact_size(self.ssl_request_bytes)

    def test_compose(self):
        self.assertEqual(self.ssl_request.compose(), self.ssl_request_bytes)


class TestSync(unittest.TestCase):
    def setUp(self):
        self.ssl_request = Sync()
        self.ssl_request_bytes = b'S'

    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            Sync.parse_exact_size(b'X')
        self.assertEqual(context_manager.exception.value, b'X')

    def test_parse(self):
        Sync.parse_exact_size(self.ssl_request_bytes)

    def test_compose(self):
        self.assertEqual(self.ssl_request.compose(), self.ssl_request_bytes)



# SPDX-License-Identifier: MPL-2.0
import abc
import enum
import typing

import attr

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import InvalidType
from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary


@attr.s
class DataAttributeBase(ParsableBase):
    """Data attribute base parser.

    .. code-block:: text

                              1                   2                   3
          0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
         !A!       Attribute Type        !    AF=0  Attribute Length     !
         !F!                             !    AF=1  Attribute Value      !
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
         .                   AF=0  Attribute Value                       .
         .                   AF=1  Not Transmitted                       .
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_type(cls):
        raise NotImplementedError()

    @classmethod
    def _get_format(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_header(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('format', 1, DataAttributeFormat)
        if parser['format'] != cls._get_format():
            raise InvalidType()

        parser.parse_numeric_enum_coded('type', type(cls.get_type()))
        if parser['type'] != cls.get_type():
            raise InvalidType()

        return parser

    def _compose_header(self):
        composer = ComposerBinary()

        composer.compose_numeric(self._get_format().value, 1)
        composer.compose_numeric_enum_coded(self.get_type())

        return composer


class DataAttributeFormat(enum.IntEnum):
    """Data attribute types."""
    TYPE_LENGTH_VALUE = 0x00
    TYPE_VALUE = 0x80


@attr.s
class DataAttributeTypeValue(DataAttributeBase):
    """Data attribute type/value parser."""

    @classmethod
    @abc.abstractmethod
    def get_type(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    def _get_format(cls):
        return DataAttributeFormat.TYPE_VALUE


@attr.s
class DataAttributeTypeValueEnumCoded(DataAttributeTypeValue):
    """Data attribute type/value parser where the type is an enum."""

    value: typing.Any = attr.ib()

    @classmethod
    @abc.abstractmethod
    def get_type(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_enum_type(cls):
        raise NotImplementedError()

    @value.validator
    def _validate_value(self, _, value):
        enum_type = self._get_enum_type()
        if not isinstance(value, enum_type):
            raise InvalidValue(value, type(self), 'value')

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_numeric_enum_coded('value', cls._get_enum_type())

        return cls(value=parser['value']), parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        composer.compose_numeric_enum_coded(self.value)

        return composer.composed_bytes


@attr.s
class DataAttributeKeyLength(DataAttributeTypeValue):
    """Key Length transform attribute (TV format).

    The Key Length attribute has the following format:

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |A|       Attribute Type        |    Key Length (in bits)       |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar key_length: Key length in bits
    """

    value: int = attr.ib(validator=attr.validators.and_(
        attr.validators.instance_of(int),
        attr.validators.ge(0),
        attr.validators.lt(2**64)
    ))

    @classmethod
    @abc.abstractmethod
    def get_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        parser.parse_numeric('value', 2)

        return cls(value=parser['value']), parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        composer.compose_numeric(self.value, 2)

        return composer.composed_bytes


@attr.s
class DataAttributeLength(DataAttributeBase):
    """Length transform attribute (TLV format).

    The Key Length attribute has the following format:

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |A|       Attribute Type        |    Key Length (in bits)       |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar key_length: Key length in bits
    """

    value: int = attr.ib(validator=attr.validators.and_(
        attr.validators.instance_of(int),
        attr.validators.ge(0),
        attr.validators.lt(2**64)
    ))

    @classmethod
    @abc.abstractmethod
    def get_type(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_size(cls):
        raise NotImplementedError()

    @classmethod
    def _get_format(cls):
        return DataAttributeFormat.TYPE_VALUE

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        if parser['format'] == DataAttributeFormat.TYPE_LENGTH_VALUE:
            parser.parse_numeric('size', 2)
            parser.parse_numeric('value', parser['size'])
        else:
            parser.parse_numeric('value', cls._get_size())

        return cls(value=parser['value']), parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        if self._get_format() == DataAttributeFormat.TYPE_LENGTH_VALUE:
            composer.compose_numeric(self._get_size(), 2)
            composer.compose_numeric(self.value, self._get_size())
        else:
            composer.compose_numeric(self.value, self._get_size())

        return composer.composed_bytes



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc
import enum

import attr

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.base import Serializable, OneByteEnumComposer, OneByteEnumParsable
from cryptoparser.common.parse import ByteOrder, ComposerBinary, ParsableBase, ParserBinary
from cryptoparser.common.exception import NotEnoughData


class MySQLVersion(enum.IntEnum):
    MYSQL_9 = 0x09
    MYSQL_10 = 0x0a


class MySQLCapability(enum.IntEnum):
    CLIENT_LONG_PASSWORD = 0x00000001
    CLIENT_FOUND_ROWS = 0x00000002
    CLIENT_LONG_FLAG = 0x00000004
    CLIENT_CONNECT_WITH_DB = 0x00000008
    CLIENT_NO_SCHEMA = 0x00000010
    CLIENT_COMPRESS = 0x00000020
    CLIENT_ODBC = 0x00000040
    CLIENT_LOCAL_FILES = 0x00000080
    CLIENT_IGNORE_SPACE = 0x00000100
    CLIENT_PROTOCOL_41 = 0x00000200
    CLIENT_INTERACTIVE = 0x00000400
    CLIENT_SSL = 0x00000800
    CLIENT_IGNORE_SIGPIPE = 0x00001000
    CLIENT_TRANSACTIONS = 0x00002000
    CLIENT_RESERVED = 0x00004000
    CLIENT_SECURE_CONNECTION = 0x00008000
    CLIENT_MULTI_STATEMENTS = 0x00010000
    CLIENT_MULTI_RESULTS = 0x00020000
    CLIENT_PS_MULTI_RESULTS = 0x00040000
    CLIENT_PLUGIN_AUTH = 0x00080000
    CLIENT_CONNECT_ATTRS = 0x00100000
    CLIENT_PLUGIN_AUTH_LENENC_CLIENT_DATA = 0x00200000
    CLIENT_CAN_HANDLE_EXPIRED_PASSWORDS = 0x00400000
    CLIENT_SESSION_TRACK = 0x00800000
    CLIENT_DEPRECATE_EOF = 0x01000000


class MySQLStatusFlag(enum.IntEnum):
    SERVER_STATUS_IN_TRANS = 0x0001
    SERVER_STATUS_AUTOCOMMIT = 0x0002
    SERVER_MORE_RESULTS_EXISTS = 0x0008
    SERVER_STATUS_NO_GOOD_INDEX_USED = 0x0010
    SERVER_STATUS_NO_INDEX_USED = 0x0020
    SERVER_STATUS_CURSOR_EXISTS = 0x0040
    SERVER_STATUS_LAST_ROW_SENT = 0x0080
    SERVER_STATUS_DB_DROPPED = 0x0100
    SERVER_STATUS_NO_BACKSLASH_ESCAPES = 0x0200
    SERVER_STATUS_METADATA_CHANGED = 0x0400
    SERVER_QUERY_WAS_SLOW = 0x0800
    SERVER_PS_OUT_PARAMS = 0x1000
    SERVER_STATUS_IN_TRANS_READONLY = 0x2000
    SERVER_SESSION_STATE_CHANGED = 0x4000


@attr.s(frozen=True)
class MySQLCharacterSetParams(Serializable):
    code = attr.ib(validator=attr.validators.instance_of(int))
    name = attr.ib(validator=attr.validators.instance_of(str))
    collate_name = attr.ib(validator=attr.validators.instance_of(str))


class MySQLCharacterSetFactory(OneByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return MySQLCharacterSet

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class MySQLCharacterSet(OneByteEnumComposer, enum.Enum):
    BIG5 = MySQLCharacterSetParams(
        code=1,
        name='big5',
        collate_name='big5_chinese_ci',
    )
    DEC8 = MySQLCharacterSetParams(
        code=3,
        name='dec8',
        collate_name='dec8_swedish_ci',
    )
    CP850 = MySQLCharacterSetParams(
        code=4,
        name='cp850',
        collate_name='cp850_general_ci',
    )
    HP8 = MySQLCharacterSetParams(
        code=6,
        name='hp8',
        collate_name='hp8_english_ci',
    )
    KOI8R = MySQLCharacterSetParams(
        code=7,
        name='koi8r',
        collate_name='koi8r_general_ci',
    )
    LATIN1 = MySQLCharacterSetParams(
        code=8,
        name='latin1',
        collate_name='latin1_swedish_ci',
    )
    LATIN2 = MySQLCharacterSetParams(
        code=9,
        name='latin2',
        collate_name='latin2_general_ci',
    )
    SWE7 = MySQLCharacterSetParams(
        code=10,
        name='swe7',
        collate_name='swe7_swedish_ci',
    )
    ASCII = MySQLCharacterSetParams(
        code=11,
        name='ascii',
        collate_name='ascii_general_ci',
    )
    UJIS = MySQLCharacterSetParams(
        code=12,
        name='ujis',
        collate_name='ujis_japanese_ci',
    )
    SJIS = MySQLCharacterSetParams(
        code=13,
        name='sjis',
        collate_name='sjis_japanese_ci',
    )
    HEBREW = MySQLCharacterSetParams(
        code=16,
        name='hebrew',
        collate_name='hebrew_general_ci',
    )
    TIS620 = MySQLCharacterSetParams(
        code=18,
        name='tis620',
        collate_name='tis620_thai_ci',
    )
    EUCKR = MySQLCharacterSetParams(
        code=19,
        name='euckr',
        collate_name='euckr_korean_ci',
    )
    KOI8U = MySQLCharacterSetParams(
        code=22,
        name='koi8u',
        collate_name='koi8u_general_ci',
    )
    GB2312 = MySQLCharacterSetParams(
        code=24,
        name='gb2312',
        collate_name='gb2312_chinese_ci',
    )
    GREEK = MySQLCharacterSetParams(
        code=25,
        name='greek',
        collate_name='greek_general_ci',
    )
    CP1250 = MySQLCharacterSetParams(
        code=26,
        name='cp1250',
        collate_name='cp1250_general_ci',
    )
    GBK = MySQLCharacterSetParams(
        code=28,
        name='gbk',
        collate_name='gbk_chinese_ci',
    )
    LATIN5 = MySQLCharacterSetParams(
        code=30,
        name='latin5',
        collate_name='latin5_turkish_ci',
    )
    ARMSCII8 = MySQLCharacterSetParams(
        code=32,
        name='armscii8',
        collate_name='armscii8_general_ci',
    )
    UTF8 = MySQLCharacterSetParams(
        code=33,
        name='utf8',
        collate_name='utf8_general_ci',
    )
    UCS2 = MySQLCharacterSetParams(
        code=35,
        name='ucs2',
        collate_name='ucs2_general_ci',
    )
    CP866 = MySQLCharacterSetParams(
        code=36,
        name='cp866',
        collate_name='cp866_general_ci',
    )
    KEYBCS2 = MySQLCharacterSetParams(
        code=37,
        name='keybcs2',
        collate_name='keybcs2_general_ci',
    )
    MACCE = MySQLCharacterSetParams(
        code=38,
        name='macce',
        collate_name='macce_general_ci',
    )
    MACROMAN = MySQLCharacterSetParams(
        code=39,
        name='macroman',
        collate_name='macroman_general_ci',
    )
    CP852 = MySQLCharacterSetParams(
        code=40,
        name='cp852',
        collate_name='cp852_general_ci',
    )
    LATIN7 = MySQLCharacterSetParams(
        code=41,
        name='latin7',
        collate_name='latin7_general_ci',
    )
    CP1251 = MySQLCharacterSetParams(
        code=51,
        name='cp1251',
        collate_name='cp1251_general_ci',
    )
    UTF16 = MySQLCharacterSetParams(
        code=54,
        name='utf16',
        collate_name='utf16_general_ci',
    )
    UTF16LE = MySQLCharacterSetParams(
        code=56,
        name='utf16le',
        collate_name='utf16le_general_ci',
    )
    CP1256 = MySQLCharacterSetParams(
        code=57,
        name='cp1256',
        collate_name='cp1256_general_ci',
    )
    CP1257 = MySQLCharacterSetParams(
        code=59,
        name='cp1257',
        collate_name='cp1257_general_ci',
    )
    UTF32 = MySQLCharacterSetParams(
        code=60,
        name='utf32',
        collate_name='utf32_general_ci',
    )
    BINARY = MySQLCharacterSetParams(
        code=63,
        name='binary',
        collate_name='binary',
    )
    GEOSTD8 = MySQLCharacterSetParams(
        code=92,
        name='geostd8',
        collate_name='geostd8_general_ci',
    )
    CP932 = MySQLCharacterSetParams(
        code=95,
        name='cp932',
        collate_name='cp932_japanese_ci',
    )
    EUCJPMS = MySQLCharacterSetParams(
        code=97,
        name='eucjpms',
        collate_name='eucjpms_japanese_ci',
    )
    GB18030 = MySQLCharacterSetParams(
        code=248,
        name='gb18030',
        collate_name='gb18030_chinese_ci',
    )
    UTF8MB4 = MySQLCharacterSetParams(
        code=255,
        name='utf8mb4',
        collate_name='utf8mb4_0900_ai_ci',
    )


class MySQLPacketBase(ParsableBase):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


@attr.s
class MySQLRecord(ParsableBase):
    HEADER_SIZE = 4

    packet_number = attr.ib(validator=attr.validators.instance_of(int))
    packet_bytes = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable, byte_order=ByteOrder.LITTLE_ENDIAN)

        parser.parse_numeric('packet_length', 3)
        parser.parse_numeric('packet_number', 1)
        parser.parse_raw('packet_bytes', parser['packet_length'])

        return MySQLRecord(
            packet_number=parser['packet_number'],
            packet_bytes=parser['packet_bytes'],
        ), parser.parsed_length

    def compose(self):
        composer = ComposerBinary(byte_order=ByteOrder.LITTLE_ENDIAN)

        composer.compose_numeric(len(self.packet_bytes), 3)
        composer.compose_numeric(self.packet_number, 1)
        composer.compose_raw(self.packet_bytes)

        return composer.composed_bytes


@attr.s
class MySQLHandshakeV10(MySQLPacketBase):  # pylint: disable=too-many-instance-attributes
    protocol_version = attr.ib(validator=attr.validators.in_(MySQLVersion))
    server_version = attr.ib(validator=attr.validators.instance_of(str))
    connection_id = attr.ib(validator=attr.validators.instance_of(int))
    auth_plugin_data = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
    capabilities = attr.ib(validator=attr.validators.deep_iterable(
        member_validator=attr.validators.instance_of(MySQLCapability),
    ))
    character_set = attr.ib(
        default=MySQLCharacterSet.UTF8, validator=attr.validators.optional(attr.validators.in_(MySQLCharacterSet))
    )
    states = attr.ib(default={}, validator=attr.validators.deep_iterable(
        member_validator=attr.validators.instance_of(MySQLStatusFlag),
    ))
    auth_plugin_data_2 = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of((bytes, bytearray))),
    )
    auth_plugin_name = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )

    MINIMUM_SIZE = 33

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.MINIMUM_SIZE:
            raise NotEnoughData(cls.MINIMUM_SIZE - len(parsable))

        parser = ParserBinary(parsable, byte_order=ByteOrder.LITTLE_ENDIAN)

        parser.parse_numeric('protocol_version', 1, MySQLVersion)
        parser.parse_string_null_terminated('server_version', 'ascii')
        parser.parse_numeric('connection_id', 4)
        parser.parse_raw('auth_plugin_data', 8)
        parser.parse_raw('filler', 1)
        del parser['filler']
        parser.parse_numeric_flags('capabilities', 2, MySQLCapability)
        parser.parse_parsable('character_set', MySQLCharacterSetFactory)
        parser.parse_numeric_flags('states', 2, MySQLStatusFlag)
        parser.parse_numeric_flags('capabilities_2', 2, MySQLCapability, shift_left=16)
        capabilities = set(parser['capabilities']) | set(parser['capabilities_2'])
        del parser['capabilities_2']

        parser.parse_numeric('auth_plugin_data_len', 1)
        auth_plugin_data_len = parser['auth_plugin_data_len']
        del parser['auth_plugin_data_len']

        parser.parse_raw('reserved', 10)
        del parser['reserved']

        if MySQLCapability.CLIENT_PLUGIN_AUTH in capabilities:
            if not auth_plugin_data_len:
                raise InvalidValue(auth_plugin_data_len, cls, 'auth_plugin_data_len')

            auth_plugin_data_2_len = auth_plugin_data_len - 8
            parser.parse_raw('auth_plugin_data_2', auth_plugin_data_2_len)

        if MySQLCapability.CLIENT_PLUGIN_AUTH in capabilities:
            parser.parse_string_null_terminated('auth_plugin_name', 'ascii')

        params = dict(parser)
        params['capabilities'] = capabilities

        return cls(**params), parser.parsed_length

    def compose(self):
        composer = ComposerBinary(byte_order=ByteOrder.LITTLE_ENDIAN)

        composer.compose_numeric(self.protocol_version, 1)
        composer.compose_string_null_terminated(self.server_version, 'ascii')
        composer.compose_numeric(self.connection_id, 4)

        composer.compose_raw(self.auth_plugin_data)

        composer.compose_raw(b'\x00')  # filler

        capabilities = [capability for capability in self.capabilities if capability.value < 2 ** 16]
        composer.compose_numeric_flags(capabilities, 2)

        capabilities_2 = [capability for capability in self.capabilities if capability.value >= 2 ** 16]
        composer.compose_parsable(self.character_set)
        composer.compose_numeric_flags(self.states, 2)
        composer.compose_numeric_flags(capabilities_2, 2, shift_right=16)

        if MySQLCapability.CLIENT_PLUGIN_AUTH in self.capabilities:
            auth_plugin_data_len = 8
            if self.auth_plugin_data_2:
                auth_plugin_data_len += len(self.auth_plugin_data_2)

            composer.compose_numeric(auth_plugin_data_len, 1)
        else:
            composer.compose_numeric(0, 1)

        composer.compose_raw(10 * b'\x00')  # reserved
        if self.auth_plugin_data_2:
            composer.compose_raw(self.auth_plugin_data_2)

        if MySQLCapability.CLIENT_PLUGIN_AUTH in self.capabilities:
            composer.compose_string_null_terminated(self.auth_plugin_name, 'ascii')

        return composer.composed_bytes


@attr.s
class MySQLHandshakeSslRequest(MySQLPacketBase):
    capabilities = attr.ib(validator=attr.validators.deep_iterable(
        member_validator=attr.validators.instance_of(MySQLCapability),
    ))
    max_packet_size = attr.ib(default=0xffff, validator=attr.validators.instance_of(int))
    character_set = attr.ib(default=None, validator=attr.validators.optional(attr.validators.in_(MySQLCharacterSet)))

    MINIMUM_SIZE = 5

    def __attrs_post_init__(self):
        if MySQLCapability.CLIENT_PROTOCOL_41 in self.capabilities:
            if self.character_set is None:
                self.character_set = MySQLCharacterSet.UTF8
        else:
            if self.max_packet_size >= 2 ** 24:
                raise ValueError(self.max_packet_size)
            for capability in self.capabilities:
                if capability.value >= 2 ** 16:
                    raise ValueError(self.max_packet_size)

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.MINIMUM_SIZE:
            raise NotEnoughData(cls.MINIMUM_SIZE - len(parsable))

        parser = ParserBinary(parsable, byte_order=ByteOrder.LITTLE_ENDIAN)

        parser.parse_numeric_flags('capabilities', 2, MySQLCapability)
        if MySQLCapability.CLIENT_PROTOCOL_41 in parser['capabilities']:
            parser.parse_numeric_flags('capabilities_2', 2, MySQLCapability, shift_left=16)
            parser.parse_numeric('max_packet_size', 4)
            parser.parse_parsable('character_set', MySQLCharacterSetFactory)
            parser.parse_raw('reserved', 23)
            del parser['reserved']

            character_set = parser['character_set']
            capabilities = parser['capabilities'] | parser['capabilities_2']
        else:
            parser.parse_numeric('max_packet_size', 3)
            capabilities = parser['capabilities']
            character_set = None

        return cls(capabilities, parser['max_packet_size'], character_set), parser.parsed_length

    def compose(self):
        composer = ComposerBinary(byte_order=ByteOrder.LITTLE_ENDIAN)

        if MySQLCapability.CLIENT_PROTOCOL_41 in self.capabilities:
            composer.compose_numeric_flags(self.capabilities, 4)
            composer.compose_numeric(self.max_packet_size, 4)
            composer.compose_parsable(self.character_set)
            composer.compose_raw(23 * b'\x00')
        else:
            composer.compose_numeric_flags(self.capabilities, 2)
            composer.compose_numeric(self.max_packet_size, 3)

        return composer.composed_bytes



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

from cryptoparser.common.field import (
    FieldValueComponentDateTime,
    FieldValueComponentQuotedString,
    FieldValueComponentString,
    FieldValueComponentTimeDelta,
    NameValuePair,
)


class HttpHeaderFieldValueComponent(NameValuePair):
    pass


class HttpHeaderFieldValueComponentExpires(FieldValueComponentDateTime):
    @classmethod
    def get_canonical_name(cls):
        return 'expires'


class HttpHeaderFieldValueComponentMaxAge(FieldValueComponentTimeDelta):
    @classmethod
    def get_canonical_name(cls):
        return 'max-age'

    @classmethod
    def _check_name(cls, name):
        cls._check_name_insensitive(name)


class HttpHeaderFieldValueComponentReport(FieldValueComponentString):
    @classmethod
    def get_canonical_name(cls):
        return 'report'

    @classmethod
    def _check_name(cls, name):
        cls._check_name_insensitive(name)


class HttpHeaderFieldValueComponentReportURI(FieldValueComponentQuotedString):
    @classmethod
    def get_canonical_name(cls):
        return 'report-uri'

    @classmethod
    def _check_name(cls, name):
        cls._check_name_insensitive(name)
